from __future__ import annotations

import hashlib
import tempfile
from glob import glob
from pathlib import Path
from typing import Optional, Union

from arti.fingerprints import Fingerprint
from arti.storage import InputFingerprints, Storage, StoragePartition
from arti.storage._internal import parse_spec, spec_to_wildcard


class LocalFilePartition(StoragePartition):
    path: str

    def compute_content_fingerprint(self, buffer_size: int = 1024 * 1024) -> Fingerprint:
        with open(self.path, mode="rb") as f:
            sha = hashlib.sha1()
            data = f.read(buffer_size)
            while len(data) > 0:
                sha.update(data)
                data = f.read(buffer_size)
        return Fingerprint.from_string(sha.hexdigest())


class LocalFile(Storage[LocalFilePartition]):
    # `_DEFAULT_PATH_TEMPLATE` and `rooted_at` ease testing, where we often want to just override
    # the tempdir, but keep the rest of the template. Eventually, we should introduce Resources and
    # implement a MockFS (to be used in `io.*`).
    _DEFAULT_PATH_TEMPLATE = str(
        Path("{graph_name}")
        / "{path_tags}"
        / "{names}"
        / "{partition_key_spec}"
        / "{input_fingerprint}"
        / "{name}{extension}"
    )

    path: str = str(Path(tempfile.gettempdir()) / _DEFAULT_PATH_TEMPLATE)

    def discover_partitions(
        self, input_fingerprints: InputFingerprints = InputFingerprints()
    ) -> tuple[LocalFilePartition, ...]:
        wildcard = spec_to_wildcard(self.path, self.key_types)
        paths = set(glob(wildcard))
        path_metadata = parse_spec(
            paths, spec=self.path, key_types=self.key_types, input_fingerprints=input_fingerprints
        )
        return tuple(
            self.generate_partition(input_fingerprint=input_fingerprint, keys=keys)
            for path, (input_fingerprint, keys) in path_metadata.items()
        )

    @classmethod
    def rooted_at(cls, root: Union[str, Path], path: Optional[str] = None) -> "LocalFile":
        path = path if path is not None else cls._DEFAULT_PATH_TEMPLATE
        return cls(path=str(Path(root) / path))



import logging
from graphlib import TopologicalSorter
from itertools import chain

from arti.artifacts import Artifact
from arti.backends import Backend
from arti.executors import Executor
from arti.graphs import Graph
from arti.producers import Producer
from arti.storage import InputFingerprints

# TODO: Factor this code out to reusable helpers/better homes.
#   - Perhaps a lot of the _build_producer logic can live in a few Producer methods. We
#   still want to expose each partition-to-build as a parallelizable unit of execution,
#   so maybe one Producer method kicks out all to-be-built partitions+metadata, and then
#   Executor can parallelize the `.build`.


class LocalExecutor(Executor):
    # TODO: Separate .map and .build steps so we can:
    # - add "sync" / "dry run" sort of things
    # - parallelize build
    #
    # We may still want to repeat the .map phase in the future, if we wanted to support some sort of
    # iterated or cyclic Producers (eg: first pass output feeds into second run - in that case,
    # `.map` should describe how to "converge" by returning the same outputs as a prior call).
    def _build_producer(
        self,
        graph: Graph,
        backend: Backend,
        producer: Producer,
    ) -> None:
        input_partitions = {
            name: backend.read_graph_partitions(
                graph.name, graph.get_snapshot_id(), graph.artifact_to_key[artifact], artifact
            )
            for name, artifact in producer.inputs.items()
        }
        partition_dependencies = producer.map(
            **{
                name: partitions
                for name, partitions in input_partitions.items()
                if name in producer._map_input_metadata_
            }
        )
        # TODO: Need to validate the partition_dependencies against the Producer's
        # partitioning scheme and such (basically, check user error). eg: if output is
        # not partitioned, we expect only 1 entry in partition_dependencies
        # (NotPartitioned).
        partition_input_fingerprints = InputFingerprints(
            {
                composite_key: producer.compute_input_fingerprint(dependency_partitions)
                for composite_key, dependency_partitions in partition_dependencies.items()
            }
        )
        output_artifacts = graph.producer_outputs[producer]
        # NOTE: The output partitions may be built, but not yet associated with this snapshot_id
        # (eg: raw input data changed, but no changes trickled into this specific Producer). Hence
        # we'll fetch all StoragePartitions for this Storage, filtered to the PKs and
        # input_fingerprints we've computed *are* for this Graph - and then link them to the graph.
        existing_output_partitions = {
            output: backend.read_artifact_partitions(output, partition_input_fingerprints)
            for output in output_artifacts
        }
        for artifact, partitions in existing_output_partitions.items():
            backend.write_graph_partitions(
                graph.name,
                graph.get_snapshot_id(),
                graph.artifact_to_key[artifact],
                artifact,
                partitions,
            )
        # TODO: Guarantee all outputs have the same set of identified partitions. Currently, this
        # pretends a partition is built for all outputs if _any_ are built for that partition.
        existing_output_keys = set(
            partition.keys for partition in chain.from_iterable(existing_output_partitions.values())
        )
        for output_partition_key, dependencies in partition_dependencies.items():
            if output_partition_key in existing_output_keys:
                pk_str = f" for: {dict(output_partition_key)}" if output_partition_key else "."
                logging.info(f"Skipping existing {type(producer).__name__} output{pk_str}")
                continue
            # TODO: Catch DispatchError and give a nicer error... maybe add this to our
            # @dispatch wrapper (eg: msg arg, or even fn that returns the message to
            # raise).
            arguments = {
                name: graph.read(
                    artifact=producer.inputs[name],
                    storage_partitions=partition_dependencies[output_partition_key][name],
                    # TODO: We'll probably need to update Producer._build_input_views_
                    # to hold *instances* (either objects set in Annotated or init
                    # during Producer.__init_subclass__)
                    view=view_class(),
                )
                for name, view_class in producer._build_input_views_.items()
            }
            outputs = producer.build(**arguments)
            if len(producer._output_metadata_) == 1:
                outputs = (outputs,)
            validation_passed, validation_message = producer.validate_outputs(*outputs)
            if not validation_passed:
                raise ValueError(validation_message)
            for i, output in enumerate(outputs):
                graph.write(
                    output,
                    artifact=output_artifacts[i],
                    input_fingerprint=partition_input_fingerprints[output_partition_key],
                    keys=output_partition_key,
                    view=producer._output_metadata_[i][1](),
                )

    def build(self, graph: Graph) -> None:
        # NOTE: Raw Artifacts will already be discovered and linked in the backend to this graph
        # snapshot.
        assert graph.snapshot_id is not None
        with graph.backend.connect() as backend:
            for node in TopologicalSorter(graph.dependencies).static_order():
                if isinstance(node, Artifact):
                    # TODO: Compute Statistics (if not already computed for the partition) and check
                    # Thresholds (every time, as they may be changed, dynamic, or overridden).
                    pass
                elif isinstance(node, Producer):
                    logging.info(f"Building {node}...")
                    self._build_producer(graph, backend, node)
                else:
                    raise NotImplementedError()
        logging.info("Build finished.")



__path__ = __import__("pkgutil").extend_path(__path__, __name__)

import abc
import os
from typing import Any, ClassVar, Generic, Optional, TypeVar

from pydantic import Field, validator

from arti.fingerprints import Fingerprint
from arti.formats import Format
from arti.internal.models import Model
from arti.internal.type_hints import get_class_type_vars, lenient_issubclass
from arti.internal.utils import frozendict
from arti.partitions import CompositeKey, CompositeKeyTypes, PartitionKey
from arti.storage._internal import InputFingerprints as InputFingerprints
from arti.storage._internal import partial_format, strip_partition_indexes
from arti.types import Type


class _StorageMixin(Model):
    @property
    def key_types(self) -> CompositeKeyTypes:
        if self.type is None:  # type: ignore
            raise ValueError(f"{self}.type is not set")
        return PartitionKey.types_from(self.type)  # type: ignore

    @classmethod
    def _check_keys(cls, key_types: CompositeKeyTypes, keys: CompositeKey) -> None:
        # TODO: Confirm the key names and types align
        if key_types and not keys:
            raise ValueError(f"Expected partition keys {tuple(key_types)} but none were passed")
        if keys and not key_types:
            raise ValueError(f"Expected no partition keys but got: {keys}")


class StoragePartition(_StorageMixin, Model):
    type: Type = Field(repr=False)
    format: Format = Field(repr=False)
    keys: CompositeKey = CompositeKey()
    input_fingerprint: Fingerprint = Fingerprint.empty()
    content_fingerprint: Fingerprint = Fingerprint.empty()

    @validator("keys")
    @classmethod
    def validate_keys(cls, keys: CompositeKey, values: dict[str, Any]) -> CompositeKey:
        if "type" in values:
            cls._check_keys(PartitionKey.types_from(values["type"]), keys)
        return keys

    def with_content_fingerprint(
        self: "_StoragePartition", keep_existing: bool = True
    ) -> "_StoragePartition":
        if keep_existing and not self.content_fingerprint.is_empty:
            return self
        return self.copy(update={"content_fingerprint": self.compute_content_fingerprint()})

    @abc.abstractmethod
    def compute_content_fingerprint(self) -> Fingerprint:
        raise NotImplementedError(
            "{type(self).__name__}.compute_content_fingerprint is not implemented!"
        )


_StoragePartition = TypeVar("_StoragePartition", bound=StoragePartition)


StoragePartitions = tuple[StoragePartition, ...]


class Storage(_StorageMixin, Model, Generic[_StoragePartition]):
    """Storage is a data reference identifying 1 or more partitions of data.

    Storage fields should have defaults set with placeholders for tags and partition
    keys. This allows automatic injection of the tags and partition keys for simple
    cases.
    """

    _abstract_ = True

    # These separators are used in the default resolve_* helpers to format metadata into
    # the storage fields.
    #
    # The defaults are tailored for "path"-like fields.
    key_value_sep: ClassVar[str] = "="
    partition_name_component_sep: ClassVar[str] = "_"
    segment_sep: ClassVar[str] = os.sep

    storage_partition_type: ClassVar[type[_StoragePartition]]  # type: ignore

    type: Optional[Type] = Field(None, repr=False)
    format: Optional[Format] = Field(None, repr=False)

    @validator("type")
    @classmethod
    def validate_type(cls, type_: Type) -> Type:
        # TODO: Check support for the types and partitioning on the specified field(s).
        return type_

    @validator("format")
    @classmethod
    def validate_format(cls, format: Format) -> Format:
        return format

    @classmethod
    def __init_subclass__(cls, **kwargs: Any) -> None:
        super().__init_subclass__(**kwargs)
        if cls._abstract_:
            return
        cls.storage_partition_type = get_class_type_vars(cls)[0]
        expected_field_types = {
            name: info.outer_type_
            for name, info in cls.storage_partition_type.__fields__.items()
            if name not in StoragePartition.__fields__
        }
        fields = {
            name: info.outer_type_
            for name, info in cls.__fields__.items()
            if name not in Storage.__fields__
        }
        if fields != expected_field_types:
            raise TypeError(
                f"{cls.__name__} fields must match {cls.storage_partition_type.__name__} ({expected_field_types}), got: {fields}"
            )

    @property
    def _format_fields(self) -> frozendict[str, str]:
        return frozendict(
            {
                name: value
                for name in self.__fields__
                if lenient_issubclass(type(value := getattr(self, name)), str)
            }
        )

    @abc.abstractmethod
    def discover_partitions(
        self, input_fingerprints: InputFingerprints = InputFingerprints()
    ) -> tuple[_StoragePartition, ...]:
        raise NotImplementedError()

    def generate_partition(
        self,
        keys: CompositeKey = CompositeKey(),
        input_fingerprint: Fingerprint = Fingerprint.empty(),
        with_content_fingerprint: bool = True,
    ) -> _StoragePartition:
        self._check_keys(self.key_types, keys)
        format_kwargs = dict[Any, Any](keys)
        if input_fingerprint.is_empty:
            if self.includes_input_fingerprint_template:
                raise ValueError(f"{self} requires an input_fingerprint, but none was provided")
        else:
            if not self.includes_input_fingerprint_template:
                raise ValueError(f"{self} does not specify a {{input_fingerprint}} template")
            format_kwargs["input_fingerprint"] = str(input_fingerprint.key)
        field_values = {
            name: (
                strip_partition_indexes(original).format(**format_kwargs)
                if lenient_issubclass(type(original := getattr(self, name)), str)
                else original
            )
            for name in self.__fields__
            if name in self.storage_partition_type.__fields__
        }
        partition = self.storage_partition_type(
            input_fingerprint=input_fingerprint, keys=keys, **field_values
        )
        if with_content_fingerprint:
            partition = partition.with_content_fingerprint()
        return partition

    @property
    def includes_input_fingerprint_template(self) -> bool:
        return any("{input_fingerprint}" in val for val in self._format_fields.values())

    def _resolve_field(self, name: str, spec: str, placeholder_values: dict[str, str]) -> str:
        for placeholder, value in placeholder_values.items():
            if not value:
                # Strip placeholder *and* any trailing self.segment_sep.
                trim = "{" + placeholder + "}"
                if f"{trim}{self.segment_sep}" in spec:
                    trim = f"{trim}{self.segment_sep}"
                # Also strip any trailing separators, eg: if the placeholder was at the end.
                spec = spec.replace(trim, "").rstrip(self.segment_sep)
                if not spec:
                    raise ValueError(f"{self}.{name} was empty after removing unused templates")
        return partial_format(spec, **placeholder_values)

    def resolve_templates(
        self: "_Storage",
        graph_name: Optional[str] = None,
        input_fingerprint: Optional[Fingerprint] = None,
        names: Optional[tuple[str, ...]] = None,
        path_tags: Optional[frozendict[str, str]] = None,
    ) -> "_Storage":
        values = {}
        if graph_name is not None:
            values["graph_name"] = graph_name
        if input_fingerprint is not None:
            input_fingerprint_key = str(input_fingerprint.key)
            if input_fingerprint.is_empty:
                input_fingerprint_key = ""
            values["input_fingerprint"] = input_fingerprint_key
        if names is not None:
            values["name"] = names[-1] if names else ""
            values["names"] = self.segment_sep.join(names)
        if path_tags is not None:
            values["path_tags"] = self.segment_sep.join(
                f"{tag}{self.key_value_sep}{value}" for tag, value in path_tags.items()
            )
        if self.format is not None:
            values["extension"] = self.format.extension
        if self.type is not None:
            key_component_specs = {
                f"{name}{self.partition_name_component_sep}{component_name}": f"{{{name}.{component_spec}}}"
                for name, pk in self.key_types.items()
                for component_name, component_spec in pk.default_key_components.items()
            }
            values["partition_key_spec"] = self.segment_sep.join(
                f"{name}{self.key_value_sep}{spec}" for name, spec in key_component_specs.items()
            )
        return self.copy(
            update={
                name: new
                for name, original in self._format_fields.items()
                # Avoid "setting" the value if not updated to reduce pydantic repr verbosity (which
                # only shows "set" fields by default).
                if (new := self._resolve_field(name, original, values)) != original
            }
        )


AnyStorage = Storage[Any]

# mypy doesn't (yet?) support nested TypeVars[1], so mark internal as Any.
#
# 1: https://github.com/python/mypy/issues/2756
_Storage = TypeVar("_Storage", bound=AnyStorage)



__path__ = __import__("pkgutil").extend_path(__path__, __name__)

from collections import defaultdict
from collections.abc import Callable, Sequence
from functools import cached_property, wraps
from graphlib import TopologicalSorter
from types import TracebackType
from typing import TYPE_CHECKING, Any, Literal, Optional, TypeVar, Union, cast

from pydantic import Field, PrivateAttr

import arti
from arti import io
from arti.artifacts import Artifact
from arti.backends import Backend
from arti.backends.memory import MemoryBackend
from arti.fingerprints import Fingerprint
from arti.internal.models import Model
from arti.internal.utils import TypedBox, frozendict
from arti.partitions import CompositeKey
from arti.producers import Producer
from arti.storage import StoragePartition, StoragePartitions
from arti.views import View

if TYPE_CHECKING:
    from arti.executors import Executor
else:
    from arti.internal.patches import patch_TopologicalSorter_class_getitem

    patch_TopologicalSorter_class_getitem()

# TODO: Add GraphMetadata model


SEALED: Literal[True] = True
OPEN: Literal[False] = False
BOX_KWARGS = {
    status: {
        "box_dots": True,
        "default_box": status is OPEN,
        "frozen_box": status is SEALED,
    }
    for status in (OPEN, SEALED)
}

_Return = TypeVar("_Return")


def requires_sealed(fn: Callable[..., _Return]) -> Callable[..., _Return]:
    @wraps(fn)
    def check_if_sealed(self: "Graph", *args: Any, **kwargs: Any) -> _Return:
        if self._status is not SEALED:
            raise ValueError(f"{fn.__name__} cannot be used while the Graph is still being defined")
        return fn(self, *args, **kwargs)

    return check_if_sealed


class ArtifactBox(TypedBox[str, Artifact]):
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        object.__setattr__(self, "_path", ())
        super().__init__(*args, **kwargs)

    def _TypedBox__cast_value(self, item: str, artifact: Any) -> Artifact:
        artifact = super()._TypedBox__cast_value(item, artifact)  # type: ignore
        storage_template_values = dict[str, Any]()
        if (graph := arti.context.graph) is not None:
            storage_template_values.update(
                {
                    "graph_name": graph.name,
                    "names": (*self._path, item),
                    "path_tags": graph.path_tags,
                }
            )
        # Require an {input_fingerprint} template in the Storage if this Artifact is being generated
        # by a Producer. Otherwise, strip the {input_fingerprint} template (if set) for "raw"
        # Artifacts.
        #
        # We can't validate this at Artifact instantiation because the Producer is tracked later (by
        # copying the instance and setting the `producer_output` attribute). We won't know the
        # "final" instance until assignment here to the Graph.
        if artifact.producer_output is None:
            storage_template_values["input_fingerprint"] = Fingerprint.empty()
        elif not artifact.storage.includes_input_fingerprint_template:
            raise ValueError(
                "Produced Artifacts must have a '{input_fingerprint}' template in their Storage"
            )
        return cast(
            Artifact,
            artifact.copy(
                update={"storage": artifact.storage.resolve_templates(**storage_template_values)}
            ),
        )

    def _Box__convert_and_store(self, item: str, value: Artifact) -> None:
        if isinstance(value, dict):
            super()._Box__convert_and_store(item, value)  # pylint: disable=no-member
            # TODO: Test if this works with `Box({"some": {"nested": "thing"}})`.
            # Guessing not, may need to put in an empty dict/box first, set the path,
            # and then update it.
            object.__setattr__(self[item], "_path", self._path + (item,))
        else:
            super()._Box__convert_and_store(item, value)  # pylint: disable=no-member

    def _Box__get_default(self, item: str, attr: bool = False) -> Any:
        value = super()._Box__get_default(item, attr=attr)  # type: ignore
        object.__setattr__(value, "_path", self._path + (item,))
        return value


Node = Union[Artifact, Producer]
NodeDependencies = frozendict[Node, frozenset[Node]]


class Graph(Model):
    """Graph stores a web of Artifacts connected by Producers."""

    _fingerprint_excludes_ = frozenset(["backend"])

    name: str
    backend: Backend = Field(default_factory=MemoryBackend)
    path_tags: frozendict[str, str] = frozendict()
    snapshot_id: Optional[Fingerprint] = None

    # Graph starts off sealed, but is opened within a `with Graph(...)` context
    _status: Optional[bool] = PrivateAttr(None)
    _artifacts: ArtifactBox = PrivateAttr(default_factory=lambda: ArtifactBox(**BOX_KWARGS[SEALED]))
    _artifact_to_key: frozendict[Artifact, str] = PrivateAttr(frozendict())

    def __enter__(self) -> "Graph":
        if arti.context.graph is not None:
            raise ValueError(f"Another graph is being defined: {arti.context.graph}")
        arti.context.graph = self
        self._toggle(OPEN)
        return self

    def __exit__(
        self,
        exc_type: Optional[type[BaseException]],
        exc_value: Optional[BaseException],
        exc_traceback: Optional[TracebackType],
    ) -> None:
        arti.context.graph = None
        self._toggle(SEALED)
        # Confirm the dependencies are acyclic
        TopologicalSorter(self.dependencies).prepare()

    def _toggle(self, status: bool) -> None:
        self._status = status
        self._artifacts = ArtifactBox(self.artifacts, **BOX_KWARGS[status])
        self._artifact_to_key = frozendict(
            {artifact: key for key, artifact in self.artifacts.walk()}
        )

    @property
    def artifacts(self) -> ArtifactBox:
        return self._artifacts

    @property
    def artifact_to_key(self) -> frozendict[Artifact, str]:
        return self._artifact_to_key

    @requires_sealed
    def build(self, executor: "Optional[Executor]" = None) -> "Graph":
        snapshot = self.snapshot()
        if executor is None:
            from arti.executors.local import LocalExecutor

            executor = LocalExecutor()
        executor.build(snapshot)
        return snapshot

    @requires_sealed
    def snapshot(self) -> "Graph":
        """Identify a "unique" ID for this Graph at this point in time.

        The ID aims to encode the structure of the Graph plus a _snapshot_ of the raw Artifact data
        (partition kinds and contents). Any change that would affect data should prompt an ID
        change, however changes to this ID don't directly cause data to be reproduced.

        NOTE: There is currently a gap (and thus race condition) between when the Graph ID is
        computed and when we read raw Artifacts data during Producer builds.
        """
        # TODO: Resolve and statically set all available fingerprints. Specifically, we
        # should pin the Producer.fingerprint, which may by dynamic (eg: version is a
        # Timestamp). Unbuilt Artifact (partitions) won't be fully resolved yet.
        if self.snapshot_id:
            return self
        snapshot_id, known_artifact_partitions = self.fingerprint, dict[str, StoragePartitions]()
        for node, _ in self.dependencies.items():
            snapshot_id = snapshot_id.combine(node.fingerprint)
            if isinstance(node, Artifact):
                key = self.artifact_to_key[node]
                snapshot_id = snapshot_id.combine(Fingerprint.from_string(key))
                # Include fingerprints (including content_fingerprint!) for all raw Artifact
                # partitions, triggering a graph ID change if these artifacts change out-of-band.
                #
                # TODO: Should we *also* inspect Producer.inputs for Artifacts _not_ inside this
                # Graph and inspect their contents too? I guess we'll have to handle things a bit
                # differently depending on if the external Artifacts are Produced (in an upstream
                # Graph) or not.
                if node.producer_output is None:
                    known_artifact_partitions[key] = StoragePartitions(
                        partition.with_content_fingerprint()
                        for partition in node.discover_storage_partitions()
                    )
                    if not known_artifact_partitions[key]:
                        content_str = "partitions" if node.is_partitioned else "data"
                        raise ValueError(f"No {content_str} found for `{key}`: {node}")
                    snapshot_id = snapshot_id.combine(
                        *[partition.fingerprint for partition in known_artifact_partitions[key]]
                    )
        if snapshot_id.is_empty or snapshot_id.is_identity:  # pragma: no cover
            # NOTE: This shouldn't happen unless the logic above is faulty.
            raise ValueError("Fingerprint is empty!")
        snapshot = self.copy(update={"snapshot_id": snapshot_id})
        assert snapshot.snapshot_id is not None  # mypy
        # Write the discovered partitions (if not already known) and link to this new snapshot.
        for key, partitions in known_artifact_partitions.items():
            snapshot.backend.write_artifact_and_graph_partitions(
                snapshot.artifacts[key], partitions, self.name, snapshot.snapshot_id, key
            )
        return snapshot

    def get_snapshot_id(self) -> Fingerprint:
        return cast(Fingerprint, self.snapshot().snapshot_id)

    @cached_property  # type: ignore # python/mypy#1362
    @requires_sealed
    def dependencies(self) -> NodeDependencies:
        artifact_deps = {
            artifact: (
                frozenset({artifact.producer_output.producer})
                if artifact.producer_output is not None
                else frozenset()
            )
            for _, artifact in self.artifacts.walk()
        }
        producer_deps = {
            # NOTE: multi-output Producers will appear multiple times (but be deduped)
            producer_output.producer: frozenset(producer_output.producer.inputs.values())
            for artifact in artifact_deps
            if (producer_output := artifact.producer_output) is not None
        }
        return NodeDependencies(artifact_deps | producer_deps)

    @cached_property  # type: ignore # python/mypy#1362
    @requires_sealed
    def producers(self) -> frozenset[Producer]:
        return frozenset(self.producer_outputs)

    @cached_property  # type: ignore # python/mypy#1362
    @requires_sealed
    def producer_outputs(self) -> frozendict[Producer, tuple[Artifact, ...]]:
        d = defaultdict[Producer, dict[int, Artifact]](dict)
        for _, artifact in self.artifacts.walk():
            if artifact.producer_output is None:
                continue
            output = artifact.producer_output
            d[output.producer][output.position] = artifact
        return frozendict(
            (producer, tuple(artifacts_by_position[i] for i in sorted(artifacts_by_position)))
            for producer, artifacts_by_position in d.items()
        )

    @requires_sealed
    def tag(self, tag: str, overwrite: bool = False) -> "Graph":
        snapshot = self.snapshot()
        assert snapshot.snapshot_id is not None
        snapshot.backend.write_graph_tag(snapshot.name, snapshot.snapshot_id, tag, overwrite)
        return snapshot

    @requires_sealed
    def from_tag(self, tag: str) -> "Graph":
        return self.copy(update={"snapshot_id": self.backend.read_graph_tag(self.name, tag)})

    # TODO: io.read/write probably need a bit of sanity checking (probably somewhere
    # else), eg: type ~= view. Doing validation on the data, etc. Should some of this
    # live on the View?

    @requires_sealed
    def read(
        self,
        artifact: Artifact,
        *,
        annotation: Optional[Any] = None,
        storage_partitions: Optional[Sequence[StoragePartition]] = None,
        view: Optional[View] = None,
    ) -> Any:
        key = self.artifact_to_key[artifact]
        if annotation is None and view is None:
            raise ValueError("Either `annotation` or `view` must be passed")
        elif annotation is not None and view is not None:
            raise ValueError("Only one of `annotation` or `view` may be passed")
        elif annotation is not None:
            view = View.get_class_for(annotation, validation_type=artifact.type)()
        assert view is not None  # mypy gets mixed up with ^
        if storage_partitions is None:
            with self.backend.connect() as backend:
                storage_partitions = backend.read_graph_partitions(
                    self.name, self.get_snapshot_id(), key, artifact
                )
        return io.read(
            type_=artifact.type,
            format=artifact.format,
            storage_partitions=storage_partitions,
            view=view,
        )

    @requires_sealed
    def write(
        self,
        data: Any,
        *,
        artifact: Artifact,
        input_fingerprint: Fingerprint = Fingerprint.empty(),
        keys: CompositeKey = CompositeKey(),
        view: Optional[View] = None,
    ) -> StoragePartition:
        key = self.artifact_to_key[artifact]
        if self.snapshot_id is not None and artifact.producer_output is None:
            raise ValueError(
                f"Writing to a raw Artifact (`{key}`) would cause a `snapshot_id` change."
            )
        if view is None:
            view = View.get_class_for(type(data), validation_type=artifact.type)()
        storage_partition = artifact.storage.generate_partition(
            input_fingerprint=input_fingerprint, keys=keys, with_content_fingerprint=False
        )
        storage_partition = io.write(
            data,
            type_=artifact.type,
            format=artifact.format,
            storage_partition=storage_partition,
            view=view,
        ).with_content_fingerprint()
        # TODO: Should we only do this in bulk? We might want the backends to
        # transparently batch requests, but that's not so friendly with the transient
        # ".connect".
        with self.backend.connect() as backend:
            backend.write_artifact_partitions(artifact, (storage_partition,))
            # Skip linking this partition to the snapshot if the id would change:
            # - If snapshot_id is already set, we'd link to the wrong snapshot (we guard against
            #   this above)
            # - If unset, we'd calculate the new id, but future `.snapshot` calls would handle too
            #   - Additionally, snapshotting may fail if not all other inputs are available now
            if artifact.producer_output is not None:
                backend.write_graph_partitions(
                    self.name, self.get_snapshot_id(), key, artifact, (storage_partition,)
                )
        return cast(StoragePartition, storage_partition)



import json
from collections.abc import Sequence
from itertools import chain
from typing import Any

from arti.formats.json import JSON
from arti.io import register_reader, register_writer
from arti.storage.literal import StringLiteralPartition, _not_written_err
from arti.types import Collection, Type
from arti.views.python import PythonBuiltin


def _read_json_literal(partition: StringLiteralPartition) -> Any:
    if partition.value is None:
        raise _not_written_err
    return json.loads(partition.value)


@register_reader
def _read_json_stringliteral_python(
    type_: Type,
    format: JSON,
    storage_partitions: Sequence[StringLiteralPartition],
    view: PythonBuiltin,
) -> Any:
    if isinstance(type_, Collection) and type_.is_partitioned:
        return list(
            chain.from_iterable(
                _read_json_literal(storage_partition) for storage_partition in storage_partitions
            )
        )
    else:
        assert len(storage_partitions) == 1  # Better error handled in base read
        return _read_json_literal(storage_partitions[0])


@register_writer
def _write_json_stringliteral_python(
    data: Any,
    type_: Type,
    format: JSON,
    storage_partition: StringLiteralPartition,
    view: PythonBuiltin,
) -> StringLiteralPartition:
    if storage_partition.value is not None:
        # We can't overwrite the original value stored in LiteralStorage - on subsequent
        # `.discover_partitions`, a partition with the original value will still be used.
        raise ValueError("Literals with a value already set cannot be written")
    return storage_partition.copy(update={"value": json.dumps(data)})



from typing import Optional

from arti.fingerprints import Fingerprint
from arti.partitions import CompositeKey
from arti.storage import InputFingerprints, Storage, StoragePartition

_not_written_err = FileNotFoundError("Literal has not been written yet")


class StringLiteralPartition(StoragePartition):
    id: str
    value: Optional[str]

    def compute_content_fingerprint(self) -> Fingerprint:
        if self.value is None:
            raise _not_written_err
        return Fingerprint.from_string(self.value)


class StringLiteral(Storage[StringLiteralPartition]):
    """StringLiteral stores a literal String value directly in the Backend."""

    id: str = "{graph_name}/{path_tags}/{names}/{partition_key_spec}/{input_fingerprint}/{name}{extension}"
    value: Optional[str]

    def discover_partitions(
        self, input_fingerprints: InputFingerprints = InputFingerprints()
    ) -> tuple[StringLiteralPartition, ...]:
        if input_fingerprints and self.value is not None:
            raise ValueError(
                f"Literal storage cannot have a `value` preset ({self.value}) for a Producer output"
            )
        if self.key_types and not input_fingerprints:
            # We won't know what partitions to lookup.
            raise ValueError("Literal storage can only be partitioned if generated by a Producer.")
        # Existing StringLiteralPartitions may be stored in the Graph's backend, however we don't
        # have access here to lookup.
        if self.value is None:
            return ()
        return tuple(
            self.generate_partition(input_fingerprint=input_fingerprint, keys=keys)
            for keys, input_fingerprint in (
                input_fingerprints or {CompositeKey(): Fingerprint.empty()}
            ).items()
        )



__path__ = __import__("pkgutil").extend_path(__path__, __name__)

from collections.abc import Sequence
from types import ModuleType
from typing import Any, Optional

from arti.formats import Format
from arti.internal.utils import dispatch, import_submodules
from arti.storage import StoragePartition, _StoragePartition
from arti.types import Collection, Type
from arti.views import View

_submodules: Optional[dict[str, ModuleType]] = None


def _discover() -> None:
    global _submodules
    if _submodules is None:
        _submodules = import_submodules(__path__, __name__)


@dispatch
def _read(
    type_: Type, format: Format, storage_partitions: Sequence[StoragePartition], view: View
) -> Any:
    raise NotImplementedError(
        f"Reading {type(storage_partitions[0])} storage in {type(format)} format to {type(view)} view is not implemented."
    )


register_reader = _read.register


def read(
    type_: Type, format: Format, storage_partitions: Sequence[StoragePartition], view: View
) -> Any:
    if not storage_partitions:
        # NOTE: Aside from simplifying this check up front, multiple dispatch with unknown list
        # element type can be ambiguous/error.
        raise FileNotFoundError("No data")
    if len(storage_partitions) > 1 and not (isinstance(type_, Collection) and type_.is_partitioned):
        raise ValueError(
            f"Multiple partitions can only be read into a partitioned Collection, not {type_}"
        )
    _discover()
    # TODO Checks that the returned data matches the Type/View
    #
    # Likely add a View method that can handle this type + schema checking, filtering to column/row subsets if necessary, etc
    return _read(type_, format, storage_partitions, view)


@dispatch  # type: ignore
def _write(
    data: Any, type_: Type, format: Format, storage_partition: _StoragePartition, view: View
) -> Optional[_StoragePartition]:
    raise NotImplementedError(
        f"Writing {type(view)} view into {type(format)} format in {type(storage_partition)} storage is not implemented."
    )


register_writer = _write.register


def write(
    data: Any, type_: Type, format: Format, storage_partition: _StoragePartition, view: View
) -> _StoragePartition:
    _discover()
    if (updated := _write(data, type_, format, storage_partition, view)) is not None:
        return updated
    return storage_partition



def patch_TopologicalSorter_class_getitem() -> None:
    """Patch adding TopologicalSorter.__class_getitem__ to support subscription.

    TopologicalSorter is considered Generic in typeshed (hence mypy expects a type arg),
    but is not at runtime.

    Pending: https://github.com/python/cpython/pull/28714
    """
    from graphlib import TopologicalSorter
    from types import GenericAlias

    if not hasattr(TopologicalSorter, "__class_getitem__"):  # pragma: no cover
        TopologicalSorter.__class_getitem__ = classmethod(GenericAlias)  # type: ignore



""" Vendored copy of find_namespace_packages from setuptools[1].

Adding setuptools as a package dependency can be a bit messy, sometimes triggering an
uninstall/reinstall on the _user_ side or other issues.

Type hints have been added to pass `mypy` strict mode.

1: https://github.com/pypa/setuptools/tree/c65e3380b0a18c92a0fc2d2b770b17cfaaec054b
"""
import os
from collections.abc import Callable, Generator, Iterable
from fnmatch import fnmatchcase


# From: https://github.com/pypa/setuptools/blob/c65e3380b0a18c92a0fc2d2b770b17cfaaec054b/setuptools/_distutils/util.py#L166
def convert_path(pathname: str) -> str:  # pragma: no cover
    """Return 'pathname' as a name that will work on the native filesystem,
    i.e. split it on '/' and put it back together again using the current
    directory separator.  Needed because filenames in the setup script are
    always supplied in Unix style, and have to be converted to the local
    convention before we can actually use them in the filesystem.  Raises
    ValueError on non-Unix-ish systems if 'pathname' either starts or
    ends with a slash.
    """
    if os.sep == "/":
        return pathname
    if not pathname:
        return pathname
    if pathname[0] == "/":
        raise ValueError("path '%s' cannot be absolute" % pathname)
    if pathname[-1] == "/":
        raise ValueError("path '%s' cannot end with '/'" % pathname)

    paths = pathname.split("/")
    while "." in paths:
        paths.remove(".")
    if not paths:
        return os.curdir
    return os.path.join(*paths)


# From: https://github.com/pypa/setuptools/blob/c65e3380b0a18c92a0fc2d2b770b17cfaaec054b/setuptools/__init__.py#L39-L119
class PackageFinder:
    """
    Generate a list of all Python packages found within a directory
    """

    @classmethod
    def find(
        cls, where: str = ".", exclude: Iterable[str] = (), include: Iterable[str] = ("*",)
    ) -> list[str]:
        """Return a list all Python packages found within directory 'where'
        'where' is the root directory which will be searched for packages.  It
        should be supplied as a "cross-platform" (i.e. URL-style) path; it will
        be converted to the appropriate local path syntax.
        'exclude' is a sequence of package names to exclude; '*' can be used
        as a wildcard in the names, such that 'foo.*' will exclude all
        subpackages of 'foo' (but not 'foo' itself).
        'include' is a sequence of package names to include.  If it's
        specified, only the named packages will be included.  If it's not
        specified, all found packages will be included.  'include' can contain
        shell style wildcard patterns just like 'exclude'.
        """

        return list(
            cls._find_packages_iter(
                convert_path(where),
                cls._build_filter("ez_setup", "*__pycache__", *exclude),
                cls._build_filter(*include),
            )
        )

    @classmethod
    def _find_packages_iter(
        cls, where: str, exclude: Callable[[str], bool], include: Callable[[str], bool]
    ) -> Generator[str, None, None]:
        """
        All the packages found in 'where' that pass the 'include' filter, but
        not the 'exclude' filter.
        """
        for root, dirs, files in os.walk(where, followlinks=True):
            # Copy dirs to iterate over it, then empty dirs.
            all_dirs = dirs[:]
            dirs[:] = []

            for dir in all_dirs:
                full_path = os.path.join(root, dir)
                rel_path = os.path.relpath(full_path, where)
                package = rel_path.replace(os.path.sep, ".")

                # Skip directory trees that are not valid packages
                if "." in dir or not cls._looks_like_package(full_path):
                    continue

                # Should this package be included?
                if include(package) and not exclude(package):
                    yield package

                # Keep searching subdirectories, as there may be more packages
                # down there, even if the parent was excluded.
                dirs.append(dir)

    @staticmethod
    def _looks_like_package(path: str) -> bool:
        """Does a directory look like a package?"""
        return os.path.isfile(os.path.join(path, "__init__.py"))

    @staticmethod
    def _build_filter(*patterns: str) -> Callable[[str], bool]:
        """
        Given a list of patterns, return a callable that will be true only if
        the input matches at least one of the patterns.
        """
        return lambda name: any(fnmatchcase(name, pat=pat) for pat in patterns)


class PEP420PackageFinder(PackageFinder):
    @staticmethod
    def _looks_like_package(path: str) -> bool:
        return True


find_packages = PackageFinder.find
find_namespace_packages = PEP420PackageFinder.find



from collections.abc import Generator, Mapping, Sequence
from copy import deepcopy
from functools import cached_property
from typing import (
    TYPE_CHECKING,
    Annotated,
    Any,
    ClassVar,
    Literal,
    Optional,
    TypeVar,
    get_args,
    get_origin,
)

from pydantic import BaseModel, Extra, root_validator, validator
from pydantic.fields import ModelField, Undefined
from pydantic.json import pydantic_encoder as pydantic_json_encoder

from arti.internal.type_hints import is_union, lenient_issubclass
from arti.internal.utils import class_name, classproperty, frozendict

if TYPE_CHECKING:
    from arti.fingerprints import Fingerprint
    from arti.types import Type


def _check_types(value: Any, type_: type) -> Any:  # noqa: C901
    mismatch_error = ValueError(f"expected an instance of {type_}, got: {value}")

    if type_ is Any:
        return value
    origin = get_origin(type_)
    if origin is not None:
        args = get_args(type_)
        if origin is Annotated:
            return _check_types(value, args[0])
        if origin is Literal:
            return _check_types(value, type(args[0]))
        # NOTE: Optional[t] -> Union[t, NoneType]
        if is_union(origin):
            for subtype in args:
                try:
                    return _check_types(value, subtype)
                except ValueError:
                    pass
            raise mismatch_error
        if issubclass(origin, (dict, Mapping)):
            value = _check_types(value, origin)
            for k, v in value.items():
                _check_types(k, args[0])
                _check_types(v, args[1])
            return value
        # Variadic tuples will be handled below
        if issubclass(origin, tuple) and ... not in args:
            value = _check_types(value, origin)
            if len(value) != len(args):
                raise mismatch_error
            for i, subtype in enumerate(args):
                _check_types(value[i], subtype)
            return value
        for t in (tuple, list, set, frozenset, Sequence):
            if issubclass(origin, t):
                value = _check_types(value, origin)
                for subvalue in value:
                    _check_types(subvalue, args[0])
                return value
        if issubclass(origin, type):
            if not lenient_issubclass(value, args[0]):
                raise ValueError(f"expected a subclass of {args[0]}, got: {value}")
            return value
        if set(args) == {Any}:
            return _check_types(value, origin)
        raise NotImplementedError(f"Missing handler for {type_} with {value}!")
    if isinstance(value, Mapping) and not isinstance(value, frozendict):
        value = frozendict(value)
    if not lenient_issubclass(type(value), type_):
        raise mismatch_error
    return value


_Model = TypeVar("_Model", bound="Model")


class Model(BaseModel):
    # A model can be marked _abstract_ to prevent direct instantiation, such as when it is intended
    # as a base class for other models with arbitrary data. As the subclasses of an _abstract_ model
    # have unknown fields (varying per subclass), we don't have targets to mark abstract with
    # abc.ABC nor typing.Protocol. See [1] for more context.
    #
    # 1: https://github.com/artigraph/artigraph/pull/60#discussion_r669089086
    _abstract_: ClassVar[bool] = True
    _class_key_: ClassVar[str] = class_name()
    _fingerprint_excludes_: ClassVar[Optional[frozenset[str]]] = None
    _fingerprint_includes_: ClassVar[Optional[frozenset[str]]] = None

    @classmethod
    def __init_subclass__(cls, **kwargs: Any) -> None:
        super().__init_subclass__(**kwargs)
        # Default _abstract_ to False if not set explicitly on the class. __dict__ is read-only.
        setattr(cls, "_abstract_", cls.__dict__.get("_abstract_", False))
        field_names = set(cls.__fields__)
        if cls._fingerprint_excludes_ and (
            unknown_excludes := cls._fingerprint_excludes_ - field_names
        ):
            raise ValueError(f"Unknown `_fingerprint_excludes_` field(s): {unknown_excludes}")
        if cls._fingerprint_includes_ and (
            unknown_includes := cls._fingerprint_includes_ - field_names
        ):
            raise ValueError(f"Unknown `_fingerprint_includes_` field(s): {unknown_includes}")

    @root_validator(pre=True)
    @classmethod
    def _block_abstract_instance(cls, values: dict[str, Any]) -> dict[str, Any]:
        if cls._abstract_:
            raise ValueError(f"{cls} cannot be instantiated directly!")
        return values

    @validator("*", pre=True)
    @classmethod
    def _strict_types(cls, value: Any, field: ModelField) -> Any:
        """Check that the value is a stricter instance of the declared type annotation.

        Pydantic will attempt to *parse* values (eg: "5" -> 5), but we'd prefer stricter values for
        clarity and to avoid silent precision loss (eg: 5.3 -> 5).
        """
        # `field.type_` points to the *inner* type (eg: `int`->`int`; `tuple[int, ...]` -> `int`)
        # while `field.outer_type_` will (mostly) include the full spec and match the `value` we
        # received. The caveat is the `field.outer_type_` will never be wrapped in `Optional`
        # (though nested fields like `tuple[tuple[Optional[int]]]` would). Hence, we pull the
        # `field.outer_type_`, but add back the `Optional` wrapping if necessary.
        type_ = field.outer_type_
        if field.allow_none:
            type_ = Optional[type_]
        return _check_types(value, type_)

    # By default, pydantic just compares models by their dict representation, causing models of
    # different types but same fields (eg: Int8 and Int16) to be equivalent. This can be removed if
    # [1] is merged+released.
    #
    # 1: https://github.com/samuelcolvin/pydantic/pull/3066
    def __eq__(self, other: Any) -> bool:
        return self.__class__ == other.__class__ and tuple(self._iter()) == tuple(other._iter())

    # Omitting unpassed args in repr by default
    def __repr_args__(self) -> Sequence[tuple[Optional[str], Any]]:
        return [(k, v) for k, v in super().__repr_args__() if k in self.__fields_set__]

    def __str__(self) -> str:
        return repr(self)

    class Config:
        extra = Extra.forbid
        frozen = True
        keep_untouched = (cached_property, classproperty)
        smart_union = True
        validate_assignment = True  # Unused with frozen, unless that is overridden in subclass.

    def copy(self: _Model, *, deep: bool = False, validate: bool = True, **kwargs: Any) -> _Model:
        copy = super().copy(deep=deep, **kwargs)
        if validate:
            # NOTE: We set exclude_unset=False so that all existing defaulted fields are reused (as
            # is normal `.copy` behavior).
            #
            # To reduce `repr` noise, we'll reset .__fields_set__ to those of the pre-validation copy
            # (which includes those originally set + updated).
            fields_set = copy.__fields_set__
            copy = copy.validate(
                dict(copy._iter(to_dict=False, by_alias=False, exclude_unset=False))
            )
            # Use object.__setattr__ to bypass frozen model assignment errors
            object.__setattr__(copy, "__fields_set__", set(fields_set))
            # Copy over the private attributes, which are missing after validation (since we're only
            # passing the fields).
            for name in self.__private_attributes__:
                if (value := getattr(self, name, Undefined)) is not Undefined:
                    if deep:
                        value = deepcopy(value)
                    object.__setattr__(copy, name, value)
        return copy

    @staticmethod
    def _fingerprint_json_encoder(obj: Any) -> Any:
        from arti.fingerprints import Fingerprint

        if isinstance(obj, Fingerprint):
            return obj.key
        if isinstance(obj, Model):
            return obj.fingerprint
        return pydantic_json_encoder(obj)

    @property
    def fingerprint(self) -> "Fingerprint":
        from arti.fingerprints import Fingerprint

        # `.json` cannot be used, even with a custom encoder, because it calls `.dict`, which
        # converts the sub-models to dicts. Instead, we want to access `.fingerprint` (in the
        # decoder).
        data = dict(
            sorted(  # Sort to ensure stability
                self._iter(
                    exclude=self._fingerprint_excludes_,
                    include=self._fingerprint_includes_,
                ),
                key=lambda kv: kv[0],
            )
        )
        json_repr = self.__config__.json_dumps(
            data,
            default=self._fingerprint_json_encoder,
        )
        return Fingerprint.from_string(f"{self._class_key_}:{json_repr}")

    # Filter out non-fields from ._iter (and thus .dict, .json, etc), such as `@cached_property`
    # after access (which just gets cached in .__dict__).
    def _iter(self, *args: Any, **kwargs: Any) -> Generator[tuple[str, Any], None, None]:
        for key, value in super()._iter(*args, **kwargs):
            if key in self.__fields__:
                yield key, value

    @classmethod
    def _pydantic_type_system_post_field_conversion_hook_(
        cls, type_: "Type", *, name: str, required: bool
    ) -> "Type":
        return type_



__path__ = __import__("pkgutil").extend_path(__path__, __name__)

import json
from itertools import chain
from typing import Any, ClassVar, Optional

from pydantic import Field, validator
from pydantic.fields import ModelField

from arti.annotations import Annotation
from arti.formats import Format
from arti.internal.models import Model
from arti.internal.type_hints import get_annotation_from_value
from arti.internal.utils import classproperty
from arti.partitions import CompositeKeyTypes, PartitionKey
from arti.storage import InputFingerprints, Storage, StoragePartition
from arti.types import Type


class BaseArtifact(Model):
    """A BaseArtifact is the most basic data structure describing data in the Artigraph ecosystem.

    A BaseArtifact is comprised of three key elements:
    - type: spec of the data's structure, such as data types, nullable, etc.
    - format: the data's serialized format, such as CSV, Parquet, database native, etc.
    - storage: the data's persistent storage system, such as blob storage, database native, etc.
    """

    # NOTE: Narrow the fields that affect the fingerprint to minimize changes (which trigger
    # recompute). Importantly, avoid fingerprinting the `.producer_output` (ie: the *upstream*
    # producer) to prevent cascading fingerprint changes (Producer.fingerprint accesses the *input*
    # Artifact.fingerprints). Even so, this may still be quite sensitive.
    _fingerprint_includes_ = frozenset(["type", "format", "storage"])

    # Type *must* be set on the class and be rather static - small additions may be necessary at
    # Graph level (eg: dynamic column additions), but these should be minor. We might allow Struct
    # Types to be "open" (partial type) or "closed".
    #
    # Format and storage *should* be set with defaults on Artifact subclasses to ease most Graph
    # definitions, but will often need to be overridden at the Graph level.
    #
    # In order to override on the instance, avoid ClassVars lest mypy complains when/if we override.
    type: Type
    format: Format
    storage: Storage[Any]

    # Hide in repr to prevent showing the entire upstream graph.
    #
    # ProducerOutput is a ForwardRef/cyclic import. Quote the entire hint to force full resolution
    # during `.update_forward_refs`, rather than `Optional[ForwardRef("ProducerOutput")]`.
    producer_output: "Optional[ProducerOutput]" = Field(None, repr=False)

    # Class level alias for `type`, which must be set on (non-abstract) subclasses.
    #
    # Pydantic removes class defaults and stashes them in cls.__fields__. To ease access, we
    # automatically populate this from `type` in `__init_subclass__`.
    _type: ClassVar[Type]

    @classmethod
    def __init_subclass__(cls, **kwargs: Any) -> None:
        super().__init_subclass__(**kwargs)
        if not cls._abstract_ and cls.__fields__["type"].default is None:
            raise ValueError(f"{cls.__name__} must set `type`")
        cls._type = cls.__fields__["type"].default

    @validator("type", always=True)
    @classmethod
    def validate_type(cls, type_: Type) -> Type:
        if type_ != cls._type:
            # NOTE: We do a lot of class level validation (particularly in Producer) that relies on
            # the *class* type, such as partition key validation. It's possible we could loosen this
            # a bit by allowing *new* Struct fields, but still requiring an exact match for other
            # Types (including existing Struct fields).
            raise ValueError("overriding `type` is not supported")
        return type_

    @validator("format", always=True)
    @classmethod
    def validate_format(cls, format: Format, values: dict[str, Any]) -> Format:
        if "type" in values:
            return format.copy(update={"type": values["type"]})
        return format

    @validator("storage", always=True)
    @classmethod
    def validate_storage(cls, storage: Storage[Any], values: dict[str, Any]) -> Storage[Any]:
        return storage.copy(
            update={name: values[name] for name in ["type", "format"] if name in values}
        ).resolve_templates()

    @classproperty
    @classmethod
    def partition_key_types(cls) -> CompositeKeyTypes:
        return PartitionKey.types_from(cls._type)

    @classproperty
    @classmethod
    def is_partitioned(cls) -> bool:
        return bool(cls.partition_key_types)

    def discover_storage_partitions(
        self, input_fingerprints: InputFingerprints = InputFingerprints()
    ) -> tuple[StoragePartition, ...]:
        # TODO: Should we support calculating the input fingerprints if not passed?
        return self.storage.discover_partitions(input_fingerprints)


class Statistic(BaseArtifact):
    """A Statistic is a piece of data derived from an Artifact that can be tracked over time."""

    # TODO: Set format/storage to some "system default" that can be used across backends?

    _abstract_ = True


class Artifact(BaseArtifact):
    """An Artifact is the base structure describing an existing or generated dataset.

    An Artifact is comprised of three key elements:
    - `type`: spec of the data's structure, such as data types, nullable, etc.
    - `format`: the data's serialized format, such as CSV, Parquet, database native, etc.
    - `storage`: the data's persistent storage system, such as blob storage, database native, etc.

    In addition to the core elements, an Artifact can be tagged with additional `annotations`
    (to associate it with human knowledge) and `statistics` (to track derived characteristics
    over time).
    """

    _abstract_ = True
    # The Artifact._by_type registry is used to track Artifact classes generated from literal python
    # values. This is populated by Artifact.from_type and used in Producer class validation
    # to find a default Artifact type for un-Annotated hints (eg: `def build(i: int)`).
    _by_type: "ClassVar[dict[Type, type[Artifact]]]" = {}

    annotations: tuple[Annotation, ...] = ()
    statistics: tuple[Statistic, ...] = ()

    @validator("annotations", "statistics", always=True, pre=True)
    @classmethod
    def _merge_class_defaults(cls, value: tuple[Any, ...], field: ModelField) -> tuple[Any, ...]:
        return tuple(chain(cls.__fields__[field.name].default, value))

    @classmethod
    def cast(cls, value: Any) -> "Artifact":
        """Attempt to convert an arbitrary value to an appropriate Artifact instance.

        `Artifact.cast` is used to convert values assigned to an `Artifact.box` (such as
        `Graph.artifacts`) into an Artifact. When called with:
        - an Artifact instance, it is returned
        - a Producer instance with a single output Artifact, the output Artifact is returned
        - a Producer instance with a multiple output Artifacts, an error is raised
        - other types, an error is raised
        """
        from arti.producers import Producer

        if isinstance(value, Artifact):
            return value
        if isinstance(value, Producer):
            output_artifacts = value.out()
            if isinstance(output_artifacts, Artifact):
                return output_artifacts
            n_outputs = len(output_artifacts)
            if n_outputs == 0:  # pragma: no cover
                # TODO: "side effect" Producers: https://github.com/artigraph/artigraph/issues/11
                raise ValueError(f"{type(value).__name__} doesn't produce any Artifacts!")
            assert n_outputs > 1
            raise ValueError(
                f"{type(value).__name__} produces {len(output_artifacts)} Artifacts. Try assigning each to a new name in the Graph!"
            )
        return Artifact.for_literal(value)

    @classmethod
    def for_literal(cls, value: Any) -> "Artifact":
        from arti.formats.json import JSON
        from arti.storage.literal import StringLiteral
        from arti.types.python import python_type_system

        annotation = get_annotation_from_value(value)
        klass = cls.from_type(python_type_system.to_artigraph(annotation, hints={}))
        return klass(
            format=JSON(),
            storage=StringLiteral(value=json.dumps(value)),
        )

    @classmethod
    def from_type(cls, type_: Type) -> "type[Artifact]":
        from arti.formats.json import JSON
        from arti.storage.literal import StringLiteral

        if type_ not in cls._by_type:
            defaults: dict[str, Any] = {
                "type": type_,
                "format": JSON(),
                "storage": StringLiteral(),  # Set a default Storage instance to support easy `Producer.out` use.
            }
            cls._by_type[type_] = type(
                f"{type_.friendly_key}Artifact",
                (cls,),
                {
                    "__annotations__": {  # Preserve the looser default type hints
                        field: cls.__fields__[field].outer_type_ for field in defaults
                    },
                    **defaults,
                },
            )
        return cls._by_type[type_]


from arti.producers import ProducerOutput  # noqa: E402



__path__ = __import__("pkgutil").extend_path(__path__, __name__)

from collections import defaultdict
from collections.abc import Callable, Iterable, Iterator
from inspect import Parameter, Signature, getattr_static
from typing import (
    TYPE_CHECKING,
    Annotated,
    Any,
    ClassVar,
    Optional,
    TypeVar,
    Union,
    cast,
    get_args,
    get_origin,
)

from arti.annotations import Annotation
from arti.artifacts import Artifact, BaseArtifact, Statistic
from arti.fingerprints import Fingerprint
from arti.internal import wrap_exc
from arti.internal.models import Model
from arti.internal.type_hints import NoneType, lenient_issubclass, signature
from arti.internal.utils import frozendict, ordinal
from arti.partitions import CompositeKey, CompositeKeyTypes, NotPartitioned
from arti.storage import StoragePartitions
from arti.versions import SemVer, Version
from arti.views import View


def _commas(vals: Iterable[Any]) -> str:
    return ", ".join([str(v) for v in vals])


# TODO: Add @validator over all (build) fields checking for an io.{read,write} handler.


ArtifactViewPair = tuple[type[Artifact], type[View]]
BuildInputViews = frozendict[str, type[View]]
MapInputMetadata = frozendict[str, type[Artifact]]
OutputMetadata = tuple[ArtifactViewPair, ...]

PartitionDependencies = frozendict[CompositeKey, frozendict[str, StoragePartitions]]

MapSig = Callable[..., PartitionDependencies]
BuildSig = Callable[..., Any]
ValidateSig = Callable[..., tuple[bool, str]]

_T = TypeVar("_T")


class Producer(Model):
    """A Producer is a task that builds one or more Artifacts."""

    # User fields/methods

    annotations: tuple[Annotation, ...] = ()
    version: Version = SemVer(major=0, minor=0, patch=1)

    # The map/build/validate_outputs parameters are intended to be dynamic and set by subclasses,
    # however mypy doesn't like the "incompatible" signature on subclasses if actually defined here
    # (nor support ParamSpec yet). `map` is generated during subclassing if not set, `build` is
    # required, and `validate_outputs` defaults to no-op checks (hence is the only one with a
    # provided method).
    #
    # These must be @classmethods or @staticmethods.
    map: ClassVar[MapSig]
    build: ClassVar[BuildSig]
    if TYPE_CHECKING:
        validate_outputs: ClassVar[ValidateSig]
    else:

        @staticmethod
        def validate_outputs(*outputs: Any) -> Union[bool, tuple[bool, str]]:
            """Validate the `Producer.build` outputs, returning the status and a message.

            The validation status is applied to all outputs. If validation does not pass, the
            outputs will not be written to storage to prevent checkpointing the output. In the
            future, we may still write the data to ease debugging, but track validation status in
            the Backend (preventing downstream use).

            The arguments must not be mutated.

            The parameters must be usable with positional argument. The output of `build` will be
            passed in as it was returned, for example: `def build(...): return 1, 2` will result in
            `validate_outputs(1, 2)`.

            NOTE: `validate_outputs` is a stopgap until Statistics and Thresholds are fully implemented.
            """
            return True, "No validation performed."

    # Internal fields/methods

    _abstract_: ClassVar[bool] = True
    _fingerprint_excludes_ = frozenset(["annotations"])

    # NOTE: The following are set in __init_subclass__
    _input_artifact_types_: ClassVar[frozendict[str, type[Artifact]]]
    _build_sig_: ClassVar[Signature]
    _build_input_views_: ClassVar[BuildInputViews]
    _output_metadata_: ClassVar[OutputMetadata]
    _map_sig_: ClassVar[Signature]
    _map_input_metadata_: ClassVar[MapInputMetadata]

    @classmethod
    def __init_subclass__(cls, **kwargs: Any) -> None:
        super().__init_subclass__(**kwargs)
        if not cls._abstract_:
            with wrap_exc(ValueError, prefix=cls.__name__):
                cls._input_artifact_types_ = cls._validate_fields()
                with wrap_exc(ValueError, prefix=".build"):
                    (
                        cls._build_sig_,
                        cls._build_input_views_,
                        cls._output_metadata_,
                    ) = cls._validate_build_sig()
                with wrap_exc(ValueError, prefix=".validate_output"):
                    cls._validate_validate_output_sig()
                with wrap_exc(ValueError, prefix=".map"):
                    cls._map_sig_, cls._map_input_metadata_ = cls._validate_map_sig()
                cls._validate_no_unused_fields()

    @classmethod
    def _get_artifact_from_annotation(cls, annotation: Any) -> type[Artifact]:
        # Avoid importing non-interface modules at root
        from arti.types.python import python_type_system

        origin, args = get_origin(annotation), get_args(annotation)
        if origin is not Annotated:
            return Artifact.from_type(python_type_system.to_artigraph(annotation, hints={}))
        annotation, *hints = args
        artifacts = [hint for hint in hints if lenient_issubclass(hint, Artifact)]
        if len(artifacts) == 0:
            return Artifact.from_type(python_type_system.to_artigraph(annotation, hints={}))
        if len(artifacts) > 1:
            raise ValueError("multiple Artifacts set")
        return cast(type[Artifact], artifacts[0])

    @classmethod
    def _get_view_from_annotation(cls, annotation: Any, artifact: type[Artifact]) -> type[View]:
        wrap_msg = f"{artifact.__name__}"
        if artifact.is_partitioned:
            wrap_msg = f"partitions of {artifact.__name__}"
        with wrap_exc(ValueError, prefix=f" ({wrap_msg})"):
            return View.get_class_for(annotation, validation_type=artifact._type)

    @classmethod
    def _validate_fields(cls) -> frozendict[str, type[Artifact]]:
        # NOTE: Aside from the base producer fields, all others should be Artifacts.
        #
        # Users can set additional class attributes, but they must be properly hinted as ClassVars.
        # These won't interact with the "framework" and can't be parameters to build/map.
        artifact_fields = {k: v for k, v in cls.__fields__.items() if k not in Producer.__fields__}
        for name, field in artifact_fields.items():
            with wrap_exc(ValueError, prefix=f".{name}"):
                if not (field.default is None and field.default_factory is None and field.required):
                    raise ValueError("field must not have a default nor be Optional.")
                if not lenient_issubclass(field.outer_type_, Artifact):
                    raise ValueError(
                        f"type hint must be an Artifact subclass, got: {field.outer_type_}"
                    )
        return frozendict({name: field.outer_type_ for name, field in artifact_fields.items()})

    @classmethod
    def _validate_parameters(
        cls, sig: Signature, *, validator: Callable[[str, Parameter, type[Artifact]], _T]
    ) -> Iterator[_T]:
        if undefined_params := set(sig.parameters) - set(cls._input_artifact_types_):
            raise ValueError(
                f"the following parameter(s) must be defined as a field: {undefined_params}"
            )
        for name, param in sig.parameters.items():
            with wrap_exc(ValueError, prefix=f" {name} param"):
                if param.annotation is param.empty:
                    raise ValueError("must have a type hint.")
                if param.default is not param.empty:
                    raise ValueError("must not have a default.")
                if param.kind not in (param.POSITIONAL_OR_KEYWORD, param.KEYWORD_ONLY):
                    raise ValueError("must be usable as a keyword argument.")
                artifact = cls.__fields__[param.name].outer_type_
                yield validator(name, param, artifact)

    @classmethod
    def _validate_build_sig_return(cls, annotation: Any, *, i: int) -> ArtifactViewPair:
        with wrap_exc(ValueError, prefix=f" {ordinal(i+1)} return"):
            artifact = cls._get_artifact_from_annotation(annotation)
            return artifact, cls._get_view_from_annotation(annotation, artifact)

    @classmethod
    def _validate_build_sig(cls) -> tuple[Signature, BuildInputViews, OutputMetadata]:
        """Validate the .build method"""
        if not hasattr(cls, "build"):
            raise ValueError("must be implemented")
        if not isinstance(getattr_static(cls, "build"), (classmethod, staticmethod)):
            raise ValueError("must be a @classmethod or @staticmethod")
        build_sig = signature(cls.build, force_tuple_return=True, remove_owner=True)
        # Validate the parameters
        build_input_metadata = BuildInputViews(
            cls._validate_parameters(
                build_sig,
                validator=(
                    lambda name, param, artifact: (
                        name,
                        cls._get_view_from_annotation(param.annotation, artifact),
                    )
                ),
            )
        )
        # Validate the return definition
        return_annotation = build_sig.return_annotation
        if return_annotation is build_sig.empty:
            # TODO: "side effect" Producers: https://github.com/artigraph/artigraph/issues/11
            raise ValueError("a return value must be set with the output Artifact(s).")
        if return_annotation == (NoneType,):
            raise ValueError("missing return signature")
        output_metadata = OutputMetadata(
            cls._validate_build_sig_return(annotation, i=i)
            for i, annotation in enumerate(return_annotation)
        )
        # Validate all output Artifacts have equivalent partitioning schemes.
        #
        # We currently require the partition key type *and* name to match, but in the
        # future we might be able to extend the dependency metadata to support
        # heterogeneous names if necessary.
        artifacts_by_composite_key = defaultdict[CompositeKeyTypes, list[type[Artifact]]](list)
        for (artifact, _) in output_metadata:
            artifacts_by_composite_key[artifact.partition_key_types].append(artifact)
        if len(artifacts_by_composite_key) != 1:
            raise ValueError("all output Artifacts must have the same partitioning scheme")
        # TODO: Save off output composite_key_types

        return build_sig, build_input_metadata, output_metadata

    @classmethod
    def _validate_validate_output_sig(cls) -> None:
        build_output_types = [
            get_args(hint)[0] if get_origin(hint) is Annotated else hint
            for hint in cls._build_sig_.return_annotation
        ]
        match_build_str = f"match the `.build` return (`{build_output_types}`)"
        validate_parameters = signature(cls.validate_outputs).parameters

        def param_matches(param: Parameter, build_return: type) -> bool:
            # Skip checking non-hinted parameters to allow lambdas.
            #
            # NOTE: Parameter type hints are *contravariant* (you can't pass a "Manager" into a
            # function expecting an "Employee"), hence the lenient_issubclass has build_return as
            # the subtype and param.annotation as the supertype.
            return param.annotation is param.empty or lenient_issubclass(
                build_return, param.annotation
            )

        if (  # Allow `*args: Any` or `*args: T` for `build(...) -> tuple[T, ...]`
            len(validate_parameters) == 1
            and (param := tuple(validate_parameters.values())[0]).kind == param.VAR_POSITIONAL
        ):
            if not all(param_matches(param, output_type) for output_type in build_output_types):
                with wrap_exc(ValueError, prefix=f" {param.name} param"):
                    raise ValueError(f"type hint must be `Any` or {match_build_str}")
        else:  # Otherwise, check pairwise
            if len(validate_parameters) != len(build_output_types):
                raise ValueError(f"must {match_build_str}")
            for i, (name, param) in enumerate(validate_parameters.items()):
                with wrap_exc(ValueError, prefix=f" {name} param"):
                    if param.default is not param.empty:
                        raise ValueError("must not have a default.")
                    if param.kind not in (param.POSITIONAL_ONLY, param.POSITIONAL_OR_KEYWORD):
                        raise ValueError("must be usable as a positional argument.")
                    if not param_matches(param, (expected := build_output_types[i])):
                        raise ValueError(
                            f"type hint must match the {ordinal(i+1)} `.build` return (`{expected}`)"
                        )
        # TODO: Validate return signature?

    @classmethod
    def _validate_map_sig(cls) -> tuple[Signature, MapInputMetadata]:
        """Validate partitioned Artifacts and the .map method"""
        if not hasattr(cls, "map"):
            partitioned_outputs = [
                artifact for (artifact, view) in cls._output_metadata_ if artifact.is_partitioned
            ]
            # TODO: Add runtime checking of `map` output (ie: output aligns w/ output
            # artifacts and such).
            if partitioned_outputs:
                raise ValueError("must be implemented when the `build` outputs are partitioned")
            else:

                def map(**kwargs: StoragePartitions) -> PartitionDependencies:
                    return PartitionDependencies(
                        {NotPartitioned: {name: partitions for name, partitions in kwargs.items()}}
                    )

            # Narrow the map signature, which is validated below and used at graph build
            # time (via cls._map_input_metadata_) to determine what arguments to pass to
            # map.
            map.__signature__ = Signature(  # type: ignore
                [
                    Parameter(name=name, annotation=StoragePartitions, kind=Parameter.KEYWORD_ONLY)
                    for name, artifact in cls._input_artifact_types_.items()
                    if name in cls._build_input_views_
                ],
                return_annotation=PartitionDependencies,
            )
            cls.map = cast(MapSig, staticmethod(map))
        if not isinstance(getattr_static(cls, "map"), (classmethod, staticmethod)):
            raise ValueError("must be a @classmethod or @staticmethod")
        map_sig = signature(cls.map)

        def validate_map_param(
            name: str, param: Parameter, artifact: type[Artifact]
        ) -> tuple[str, type[Artifact]]:
            # TODO: Should we add some ArtifactPartition[MyArtifact] type?
            if param.annotation != StoragePartitions:
                raise ValueError("type hint must be `StoragePartitions`")
            return name, artifact

        map_input_metadata = MapInputMetadata(
            cls._validate_parameters(map_sig, validator=validate_map_param)
        )
        return map_sig, map_input_metadata  # TODO: Verify map output hint matches TBD spec

    @classmethod
    def _validate_no_unused_fields(cls) -> None:
        if unused_fields := set(cls._input_artifact_types_) - (
            set(cls._build_sig_.parameters) | set(cls._map_sig_.parameters)
        ):
            raise ValueError(
                f"the following fields aren't used in `.build` or `.map`: {unused_fields}"
            )

    # NOTE: pydantic defines .__iter__ to return `self.__dict__.items()` to support `dict(model)`,
    # but we want to override to support easy expansion/assignment to a Graph  without `.out()` (eg:
    # `g.artifacts.a, g.artifacts.b = MyProducer(...)`).
    def __iter__(self) -> Iterator[Artifact]:  # type: ignore
        ret = self.out()
        if not isinstance(ret, tuple):
            ret = (ret,)
        return iter(ret)

    def compute_input_fingerprint(
        self, dependency_partitions: frozendict[str, StoragePartitions]
    ) -> Fingerprint:
        input_names = set(dependency_partitions)
        expected_names = set(self._build_input_views_)
        if input_names != expected_names:
            raise ValueError(
                f"Mismatched dependency inputs; expected {expected_names}, got {input_names}"
            )
        # We only care if the *code* or *input partition contents* changed, not if the input file
        # paths changed (but have the same content as a prior run).
        return Fingerprint.from_string(self._class_key_).combine(
            self.version.fingerprint,
            *(
                partition.content_fingerprint
                for name, partitions in dependency_partitions.items()
                for partition in partitions
            ),
        )

    @property
    def inputs(self) -> dict[str, Artifact]:
        return {k: getattr(self, k) for k in self._input_artifact_types_}

    def out(self, *outputs: Artifact) -> Union[Artifact, tuple[Artifact, ...]]:
        """Configure the output Artifacts this Producer will build.

        The arguments are matched to the `Producer.build` return signature in order.
        """
        if not outputs:
            # TODO: Raise a better error if the Artifacts don't have defaults set for
            # type/format/storage.
            outputs = tuple(artifact() for (artifact, _) in self._output_metadata_)
        passed_n, expected_n = len(outputs), len(self._build_sig_.return_annotation)
        if passed_n != expected_n:
            ret_str = _commas(self._build_sig_.return_annotation)
            raise ValueError(
                f"{self._class_key_}.out() - expected {expected_n} arguments of ({ret_str}), but got: {outputs}"
            )

        def validate(artifact: Artifact, *, ord: int) -> Artifact:
            (expected_type, _) = self._output_metadata_[ord]
            with wrap_exc(ValueError, prefix=f"{self._class_key_}.out() {ordinal(ord+1)} argument"):
                if not isinstance(artifact, expected_type):
                    raise ValueError(f"expected instance of {expected_type}, got {type(artifact)}")
                # TODO: Validate type/format/storage/view compatibility?
                if artifact.producer_output is not None:
                    raise ValueError(
                        f"{artifact} is produced by {artifact.producer_output.producer}!"
                    )
            return artifact.copy(
                update={"producer_output": ProducerOutput(producer=self, position=ord)}
            )

        outputs = tuple(validate(artifact, ord=i) for i, artifact in enumerate(outputs))
        if len(outputs) == 1:
            return outputs[0]
        return outputs


def producer(
    *,
    annotations: Optional[tuple[Annotation, ...]] = None,
    map: Optional[MapSig] = None,
    name: Optional[str] = None,
    validate_outputs: Optional[ValidateSig] = None,
    version: Optional[Version] = None,
) -> Callable[[BuildSig], type[Producer]]:
    def decorate(build: BuildSig) -> type[Producer]:
        nonlocal name
        name = build.__name__ if name is None else name
        __annotations__: dict[str, Any] = {}
        for param in signature(build).parameters.values():
            with wrap_exc(ValueError, prefix=f"{name} {param.name} param"):
                __annotations__[param.name] = Producer._get_artifact_from_annotation(
                    param.annotation
                )
        # If overriding, set an explicit "annotations" hint until [1] is released.
        #
        # 1: https://github.com/samuelcolvin/pydantic/pull/3018
        if annotations:
            __annotations__["annotations"] = tuple[Annotation, ...]
        if version:
            __annotations__["version"] = Version
        return type(
            name,
            (Producer,),
            {
                k: v
                for k, v in {
                    "__annotations__": __annotations__,
                    "annotations": annotations,
                    "build": staticmethod(build),
                    "map": None if map is None else staticmethod(map),
                    "validate_outputs": (
                        None if validate_outputs is None else staticmethod(validate_outputs)
                    ),
                    "version": version,
                }.items()
                if v is not None
            },
        )

    return decorate


class ProducerOutput(Model):
    producer: Producer
    position: int  # TODO: Support named output (defaulting to artifact classname)


BaseArtifact.update_forward_refs(ProducerOutput=ProducerOutput)
Statistic.update_forward_refs(ProducerOutput=ProducerOutput)
Artifact.update_forward_refs(ProducerOutput=ProducerOutput)



__path__ = __import__("pkgutil").extend_path(__path__, __name__)

from abc import abstractmethod
from collections.abc import Iterator
from contextlib import contextmanager
from typing import TypeVar

from arti.artifacts import Artifact
from arti.fingerprints import Fingerprint
from arti.internal.models import Model
from arti.storage import InputFingerprints, StoragePartitions

# TODO: Consider adding CRUD methods for "everything"?
#
# Likely worth making a distinction between lower level ("CRUD") methods vs higher level ("RPC" or
# "composing") methods. Executors should operate on the high level methods, but those may have
# defaults simply calling the lower level methods. If high level methods can be optimized (eg: not a
# bunch of low level calls, each own network call), Backend can override.


_Backend = TypeVar("_Backend", bound="Backend")


class Backend(Model):
    """Backend represents a storage for internal Artigraph metadata.

    Backend storage is an addressable location (local path, database connection, etc) that
    tracks metadata for a collection of Graphs over time, including:
    - the Artifact(s)->Producer->Artifact(s) dependency graph
    - Artifact Annotations, Statistics, Partitions, and other metadata
    - Artifact and Producer Fingerprints
    - etc
    """

    @contextmanager
    def connect(self: _Backend) -> Iterator[_Backend]:
        raise NotImplementedError()

    # Artifact partitions - independent of a specific Graph (snapshot)

    @abstractmethod
    def read_artifact_partitions(
        self, artifact: Artifact, input_fingerprints: InputFingerprints = InputFingerprints()
    ) -> StoragePartitions:
        """Read all known Partitions for this Storage spec.

        If `input_fingerprints` is provided, the returned partitions will be filtered accordingly.

        NOTE: The returned partitions may not be associated with any particular Graph, unless
        `input_fingerprints` is provided matching those for a Graph snapshot.
        """
        raise NotImplementedError()

    @abstractmethod
    def write_artifact_partitions(self, artifact: Artifact, partitions: StoragePartitions) -> None:
        """Add more partitions for a Storage spec."""
        raise NotImplementedError()

    # Artifact partitions for a specific Graph (snapshot)

    @abstractmethod
    def read_graph_partitions(
        self, graph_name: str, graph_snapshot_id: Fingerprint, artifact_key: str, artifact: Artifact
    ) -> StoragePartitions:
        """Read the known Partitions for the named Artifact in a specific Graph snapshot."""
        raise NotImplementedError()

    @abstractmethod
    def write_graph_partitions(
        self,
        graph_name: str,
        graph_snapshot_id: Fingerprint,
        artifact_key: str,
        artifact: Artifact,
        partitions: StoragePartitions,
    ) -> None:
        """Link the Partitions to the named Artifact in a specific Graph snapshot."""
        raise NotImplementedError()

    def write_artifact_and_graph_partitions(
        self,
        artifact: Artifact,
        partitions: StoragePartitions,
        graph_name: str,
        graph_snapshot_id: Fingerprint,
        artifact_key: str,
    ) -> None:
        self.write_artifact_partitions(artifact, partitions)
        self.write_graph_partitions(
            graph_name, graph_snapshot_id, artifact_key, artifact, partitions
        )

    # Graph Snapshot Tagging

    @abstractmethod
    def read_graph_tag(self, graph_name: str, tag: str) -> Fingerprint:
        """Fetch the Snapshot ID for the named tag."""
        raise NotImplementedError()

    @abstractmethod
    def write_graph_tag(
        self, graph_name: str, graph_snapshot_id: Fingerprint, tag: str, overwrite: bool = False
    ) -> None:
        """Tag a Graph Snapshot ID with an arbitrary name."""
        raise NotImplementedError()



__path__ = __import__("pkgutil").extend_path(__path__, __name__)

import operator
from collections.abc import Callable
from functools import reduce
from typing import Optional, Union

import farmhash

from arti.internal.models import Model
from arti.internal.utils import int64, uint64


def _gen_fingerprint_binop(
    op: Callable[[int, int], int]
) -> "Callable[[Fingerprint, Union[int, Fingerprint]], Fingerprint]":
    def _fingerprint_binop(self: "Fingerprint", other: "Union[int, Fingerprint]") -> "Fingerprint":
        if isinstance(other, int):
            other = Fingerprint.from_int(other)
        if isinstance(other, Fingerprint):
            if self.key is None or other.key is None:
                return Fingerprint.empty()
            return Fingerprint(key=op(self.key, other.key))
        return NotImplemented

    return _fingerprint_binop


class Fingerprint(Model):
    """Fingerprint represents a unique identity as an int64 value.

    Using an int(64) has a number of convenient properties:
    - can be combined independent of order with XOR
    - can be stored relatively cheaply
    - empty 0 values drop out when combined (5 ^ 0 = 5)
    - is relatively cross-platform (across databases, languages, etc)

    There are two "special" Fingerprints w/ factory functions that, when combined with other
    Fingerprints:
    - `empty()`: returns `empty()`
    - `identity()`: return the other Fingerprint
    """

    key: Optional[int64]

    def combine(self, *others: "Fingerprint") -> "Fingerprint":
        return reduce(operator.xor, others, self)

    @classmethod
    def empty(cls) -> "Fingerprint":
        """Return a Fingerprint that, when combined, will return Fingerprint.empty()"""
        return cls(key=None)

    @classmethod
    def from_int(cls, x: int, /) -> "Fingerprint":
        return cls.from_int64(int64(x))

    @classmethod
    def from_int64(cls, x: int64, /) -> "Fingerprint":
        return cls(key=x)

    @classmethod
    def from_string(cls, x: str, /) -> "Fingerprint":
        """Fingerprint an arbitrary string.

        Fingerprints using Farmhash Fingerprint64, converted to int64 via two's complement.
        """
        return cls.from_uint64(uint64(farmhash.fingerprint64(x)))

    @classmethod
    def from_uint64(cls, x: uint64, /) -> "Fingerprint":
        return cls.from_int64(int64(x))

    @classmethod
    def identity(cls) -> "Fingerprint":
        """Return a Fingerprint that, when combined, will return the other Fingerprint."""
        return cls(key=int64(0))

    @property
    def is_empty(self) -> bool:
        return self.key is None

    @property
    def is_identity(self) -> bool:
        return self.key == 0

    __and__ = _gen_fingerprint_binop(operator.__and__)
    __lshift__ = _gen_fingerprint_binop(operator.__lshift__)
    __or__ = _gen_fingerprint_binop(operator.__or__)
    __rshift__ = _gen_fingerprint_binop(operator.__rshift__)
    __xor__ = _gen_fingerprint_binop(operator.__xor__)

    def __eq__(self, other: object) -> bool:
        if isinstance(other, int):
            other = Fingerprint.from_int(other)
        if isinstance(other, Fingerprint):
            return self.key == other.key
        return NotImplemented



from __future__ import annotations

from collections import defaultdict
from collections.abc import Iterator
from contextlib import contextmanager

from pydantic import PrivateAttr

from arti.artifacts import Artifact
from arti.backends import Backend
from arti.fingerprints import Fingerprint
from arti.storage import AnyStorage, InputFingerprints, StoragePartition, StoragePartitions


def ensure_fingerprinted(partitions: StoragePartitions) -> Iterator[StoragePartition]:
    for partition in partitions:
        yield partition.with_content_fingerprint(keep_existing=True)


_GraphSnapshotPartitions = dict[str, dict[Fingerprint, dict[str, set[StoragePartition]]]]
_StoragePartitions = dict[AnyStorage, set[StoragePartition]]


class MemoryBackend(Backend):
    # `_graph_snapshot_partitions` tracks all the partitions for a *specific* "run" of a graph.
    # `_storage_partitions` tracks all partitions, across all graphs. This separation is important
    # to allow for Literals to be used even after a snapshot_id change.
    _graph_snapshot_partitions: _GraphSnapshotPartitions = PrivateAttr(
        default_factory=lambda: defaultdict(
            lambda: defaultdict(lambda: defaultdict(set[StoragePartition]))
        )
    )
    _graph_tags: dict[str, dict[str, Fingerprint]] = PrivateAttr(
        default_factory=lambda: defaultdict(dict)
    )
    _storage_partitions: _StoragePartitions = PrivateAttr(
        default_factory=lambda: defaultdict(set[StoragePartition])
    )

    @contextmanager
    def connect(self) -> Iterator[MemoryBackend]:
        yield self

    def read_artifact_partitions(
        self, artifact: Artifact, input_fingerprints: InputFingerprints = InputFingerprints()
    ) -> StoragePartitions:
        # The MemoryBackend is (obviously) not persistent, so there may be external data we don't
        # know about. If we haven't seen this storage before, we'll attempt to "warm" the cache.
        if artifact.storage not in self._storage_partitions:
            self.write_artifact_partitions(
                artifact, artifact.discover_storage_partitions(input_fingerprints)
            )
        partitions = self._storage_partitions[artifact.storage]
        if input_fingerprints:
            partitions = {
                partition
                for partition in partitions
                if input_fingerprints.get(partition.keys) == partition.input_fingerprint
            }
        return tuple(partitions)

    def write_artifact_partitions(self, artifact: Artifact, partitions: StoragePartitions) -> None:
        self._storage_partitions[artifact.storage].update(ensure_fingerprinted(partitions))

    def read_graph_partitions(
        self, graph_name: str, graph_snapshot_id: Fingerprint, artifact_key: str, artifact: Artifact
    ) -> StoragePartitions:
        return tuple(self._graph_snapshot_partitions[graph_name][graph_snapshot_id][artifact_key])

    def write_graph_partitions(
        self,
        graph_name: str,
        graph_snapshot_id: Fingerprint,
        artifact_key: str,
        artifact: Artifact,
        partitions: StoragePartitions,
    ) -> None:
        self._graph_snapshot_partitions[graph_name][graph_snapshot_id][artifact_key].update(
            ensure_fingerprinted(partitions)
        )

    def read_graph_tag(self, graph_name: str, tag: str) -> Fingerprint:
        if tag not in self._graph_tags[graph_name]:
            raise ValueError(f"No known `{tag}` tag for Graph `{graph_name}`")
        return self._graph_tags[graph_name][tag]

    def write_graph_tag(
        self, graph_name: str, graph_snapshot_id: Fingerprint, tag: str, overwrite: bool = False
    ) -> None:
        """Read the known Partitions for the named Artifact in a specific Graph snapshot."""
        if (existing := self._graph_tags[graph_name].get(tag)) is not None and not overwrite:
            raise ValueError(f"Existing `{tag}` tag for Graph `{graph_name}` points to {existing}")
        self._graph_tags[graph_name][tag] = graph_snapshot_id



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['arti',
 'arti.annotations',
 'arti.artifacts',
 'arti.backends',
 'arti.executors',
 'arti.fingerprints',
 'arti.formats',
 'arti.graphs',
 'arti.internal',
 'arti.internal.vendored',
 'arti.io',
 'arti.partitions',
 'arti.producers',
 'arti.statistics',
 'arti.storage',
 'arti.thresholds',
 'arti.types',
 'arti.versions',
 'arti.views']

package_data = \
{'': ['*']}

install_requires = \
['frozendict>=2.1.0,<3.0.0',
 'multimethod>=1.6,<2.0',
 'parse>=1.19.0,<2.0.0',
 'pydantic>=1.9.0,<2.0.0',
 'pyfarmhash>=0.2.2,<0.4.0',
 'python-box>=5.4.1,<6.0.0']

setup_kwargs = {
    'name': 'arti',
    'version': '0.0.1',
    'description': '',
    'long_description': '# artigraph\n\n[![pypi](https://img.shields.io/pypi/v/arti.svg)](https://pypi.python.org/pypi/arti)\n[![downloads](https://pepy.tech/badge/arti/month)](https://pepy.tech/project/arti)\n[![versions](https://img.shields.io/pypi/pyversions/arti.svg)](https://github.com/artigraph/artigraph)\n[![license](https://img.shields.io/github/license/artigraph/artigraph.svg)](https://github.com/artigraph/artigraph/blob/golden/LICENSE)\n[![CI](https://github.com/artigraph/artigraph/actions/workflows/ci.yaml/badge.svg)](https://github.com/artigraph/artigraph/actions/workflows/ci.yaml)\n[![codecov](https://codecov.io/gh/artigraph/artigraph/branch/golden/graph/badge.svg?token=6LUCpjcGdN)](https://codecov.io/gh/artigraph/artigraph)\n\nDeclarative Data Production\n\n## Installation\n\nArtigraph can be installed from PyPI on python 3.9+ with `pip install arti`.\n\n## Example\n\nThis [simple example](docs/examples/spend/demo.py) takes a series of purchase transactions and computes the total amount spent:\n\n```python\nfrom pathlib import Path\nfrom typing import Annotated\n\nfrom arti import Annotation, Artifact, Graph, producer\nfrom arti.formats.json import JSON\nfrom arti.storage.local import LocalFile\nfrom arti.types import Collection, Date, Float64, Int64, Struct\nfrom arti.versions import SemVer\n\nDIR = Path(__file__).parent\n\n\nclass Vendor(Annotation):\n    name: str\n\n\nclass Transactions(Artifact):\n    """Transactions partitioned by day."""\n\n    type = Collection(\n        element=Struct(fields={"id": Int64(), "date": Date(), "amount": Float64()}),\n        partition_by=("date",),\n    )\n\n\nclass TotalSpend(Artifact):\n    """Aggregate spend over all time."""\n\n    type = Float64()\n    format = JSON()\n    storage = LocalFile()\n\n\n@producer(version=SemVer(major=1, minor=0, patch=0))\ndef aggregate_transactions(\n    transactions: Annotated[list[dict], Transactions]\n) -> Annotated[float, TotalSpend]:\n    return sum(txn["amount"] for txn in transactions)\n\n\nwith Graph(name="test") as g:\n    g.artifacts.vendor.transactions = Transactions(\n        annotations=[Vendor(name="Acme")],\n        format=JSON(),\n        storage=LocalFile(path=str(DIR / "transactions" / "{date.iso}.json")),\n    )\n    g.artifacts.spend = aggregate_transactions(transactions=g.artifacts.vendor.transactions)\n```\n\nThis example can be run easily with `docker run --rm artigraph/example-spend`.\n',
    'author': 'Jacob Hayes',
    'author_email': 'jacob.r.hayes@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/artigraph/artigraph',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)



__path__ = __import__("pkgutil").extend_path(__path__, __name__)
from typing import Annotated, Any, ClassVar, Optional, get_args, get_origin

from arti.internal.models import Model
from arti.internal.type_hints import lenient_issubclass
from arti.internal.utils import import_submodules, register
from arti.types import Type, TypeSystem


class View(Model):
    """View represents the in-memory representation of the artifact.

    Examples include pandas.DataFrame, dask.DataFrame, a BigQuery table.
    """

    _abstract_ = True
    _by_python_type_: "ClassVar[dict[type, type[View]]]" = {}

    priority: ClassVar[int] = 0  # Set priority of this view for its python_type. Higher is better.
    python_type: ClassVar[type]
    type_system: ClassVar[TypeSystem]

    @classmethod
    def __init_subclass__(cls, **kwargs: Any) -> None:
        super().__init_subclass__(**kwargs)
        if not cls._abstract_:
            register(cls._by_python_type_, cls.python_type, cls, lambda x: x.priority)

    @classmethod
    def get_class_for(
        cls, annotation: Any, *, validation_type: Optional[Type] = None
    ) -> type["View"]:
        import_submodules(__path__, __name__)

        view = None

        origin, args = get_origin(annotation), get_args(annotation)
        if origin is Annotated:
            annotation, *hints = args
            views: list[type[View]] = [hint for hint in hints if lenient_issubclass(hint, View)]
            if len(views) == 0:
                return cls.get_class_for(annotation, validation_type=validation_type)
            if len(views) == 1:
                view = views[0]
            else:
                raise ValueError("multiple Views set")
        if view is None:
            if origin is None:
                origin = annotation
            if origin not in cls._by_python_type_:
                raise ValueError(
                    f"{annotation} cannot be matched to a View, try setting one explicitly (eg: `Annotated[pd.DataFrame, MyArtifact, PandasDataFrame]`)"
                )
            view = cls._by_python_type_[origin]
        if validation_type is not None:
            view.check_type_similarity(arti=validation_type, python_type=annotation)
        return view

    @classmethod
    def check_type_similarity(cls, *, arti: Type, python_type: type) -> None:
        system_type = cls.type_system.to_system(arti, hints={})
        if not (
            lenient_issubclass(system_type, python_type)
            or lenient_issubclass(type(system_type), python_type)
        ):
            raise ValueError(f"{python_type} cannot be used to represent {arti}")



import json
from collections.abc import Sequence
from itertools import chain
from pathlib import Path
from typing import Any

from arti.formats.json import JSON
from arti.io import register_reader, register_writer
from arti.storage.local import LocalFilePartition
from arti.types import Collection, Type
from arti.views.python import PythonBuiltin

# TODO: Do I need to inject the partition keys into the returned data? Likely useful...
# Maybe a View option?


def _read_json_file(path: str) -> Any:
    with open(path, "r") as file:
        return json.load(file)


@register_reader
def _read_json_localfile_python(
    type_: Type,
    format: JSON,
    storage_partitions: Sequence[LocalFilePartition],
    view: PythonBuiltin,
) -> Any:
    if isinstance(type_, Collection) and type_.is_partitioned:
        return list(
            chain.from_iterable(
                _read_json_file(storage_partition.path) for storage_partition in storage_partitions
            )
        )
    else:
        assert len(storage_partitions) == 1  # Better error handled in base read
        return _read_json_file(storage_partitions[0].path)


@register_writer
def _write_json_localfile_python(
    data: Any, type_: Type, format: JSON, storage_partition: LocalFilePartition, view: PythonBuiltin
) -> None:
    path = Path(storage_partition.path)
    path.parent.mkdir(exist_ok=True, parents=True)
    with path.open("w") as file:
        json.dump(data, file)



from __future__ import annotations

__path__ = __import__("pkgutil").extend_path(__path__, __name__)

import threading
from typing import Optional

from arti.annotations import Annotation
from arti.artifacts import Artifact
from arti.backends import Backend
from arti.executors import Executor
from arti.fingerprints import Fingerprint
from arti.formats import Format
from arti.graphs import Graph
from arti.io import read, register_reader, register_writer, write
from arti.partitions import CompositeKey, CompositeKeyTypes, PartitionKey
from arti.producers import PartitionDependencies, Producer, producer
from arti.statistics import Statistic
from arti.storage import InputFingerprints, Storage, StoragePartition, StoragePartitions
from arti.thresholds import Threshold
from arti.types import Type, TypeAdapter, TypeSystem
from arti.versions import Version
from arti.views import View

# Export all interfaces.
__all__ = [
    "Annotation",
    "Artifact",
    "Backend",
    "CompositeKey",
    "CompositeKeyTypes",
    "Executor",
    "Fingerprint",
    "Format",
    "Graph",
    "InputFingerprints",
    "PartitionDependencies",
    "PartitionKey",
    "Producer",
    "Statistic",
    "Storage",
    "StoragePartition",
    "StoragePartitions",
    "Threshold",
    "Type",
    "TypeAdapter",
    "TypeSystem",
    "Version",
    "View",
    "producer",
    "read",
    "register_reader",
    "register_writer",
    "write",
]


class _Context(threading.local):
    def __init__(self) -> None:
        super().__init__()
        self.graph: Optional[Graph] = None


context = _Context()



import pickle
from collections.abc import Sequence
from itertools import chain
from pathlib import Path
from typing import Any

from arti.formats.pickle import Pickle
from arti.io import register_reader, register_writer
from arti.storage.local import LocalFilePartition
from arti.types import Collection, Type
from arti.views.python import PythonBuiltin


def _read_pickle_file(path: str) -> Any:
    with open(path, "rb") as file:
        return pickle.load(file)


@register_reader
def _read_pickle_localfile_python(
    type_: Type,
    format: Pickle,
    storage_partitions: Sequence[LocalFilePartition],
    view: PythonBuiltin,
) -> Any:
    if isinstance(type_, Collection) and type_.is_partitioned:
        return list(
            chain.from_iterable(
                _read_pickle_file(storage_partition.path)
                for storage_partition in storage_partitions
            )
        )
    else:
        assert len(storage_partitions) == 1  # Better error handled in base read
        return _read_pickle_file(storage_partitions[0].path)


@register_writer
def _write_pickle_localfile_python(
    data: Any,
    type_: Type,
    format: Pickle,
    storage_partition: LocalFilePartition,
    view: PythonBuiltin,
) -> None:
    path = Path(storage_partition.path)
    path.parent.mkdir(exist_ok=True, parents=True)
    with path.open("wb") as file:
        pickle.dump(data, file)
